import java.net.*;
import java.util.StringTokenizer;
import java.io.*;

import javax.media.Format;

import javax.swing.*;


/*
 * Created on Feb 27, 2005
*/
/** 
 * The Connection with the client is established by server
 * then ClientHandler thread is opened
 * this thread manages everything about a client
 * 1. processes connections
 * 2. Gets io strems 
 * 3. closes connections
 * * @author popel
 */

public class ClientHandler implements Runnable{
	/* (non-Javadoc)
	 * @see java.lang.Runnable#run()
	 */
	private Socket client;
	private Server myBoss;
	private ObjectInputStream clientIn;
	private ObjectOutputStream clientOut;
	private LogRecorder logClient;
	private String user;
	private String pass;
	private int portNumber;
	private ServerEventHandler clientEventHandler;
	private int currentActivity = BEING_AUTHORIZED;
	private int clientNumber;
	
	public static final int BEING_AUTHORIZED  = 0;
	public static final int IDLE = 1;
	public static final int WATCHING_VIDEO = 2;
	public static final int WATCHING_TV = 3;
	
	/*
	 *  1 means idle
	 *  2 means watching video
	 *  3 means watching tv
	 *  4 means on a video conference
	 *  5 means on a voice chat
	 *  6 means on a private session
	 */
	private AVTransmitter avTransmitter = null;
	private RTPTransmitter rtpTransmitter;
	
	public void setClientEventHandler(ServerEventHandler x)
	{
		clientEventHandler = x;
		
	}
	ClientHandler(Server server,Socket s,int clientNum)
	{
		myBoss = server;
		client = s;
		clientEventHandler = new ServerEventHandlerAdapter();
		portNumber = myBoss.myMediaManager.getAvailablePortNumber();
		//System.out.println("Client From "+client.getInetAddress().getHostAddress());
		clientNumber  =clientNum;
		
	}
	private void getStreams() throws IOException
	{
		clientOut = new ObjectOutputStream(client.getOutputStream());
		
		//System.out.println("Got Output stream "+clientOut.toString());
		clientOut.flush(); // flush output buffer.
		
		
		clientIn = new ObjectInputStream(client.getInputStream());
		//System.out.println("Got Input stream "+clientIn.toString());
		
	}
		
	private void closeConnection()
	{
		
		try{
			clientIn.close();
			clientOut.close();
			client.close();

			
			
		}
		catch(IOException ioException)
		{
			System.err.println("Error occurs during closing streams for client");
			
		}
		finally{
			if(logClient!=null)logClient.close();
			
			if(avTransmitter!=null){
				avTransmitter.stop();
				avTransmitter = null;
			}
			if(rtpTransmitter!=null){
				rtpTransmitter.stop();
				rtpTransmitter = null;
			}
			// return the port to Media Manager as it was alloted
			myBoss.myMediaManager.addPort(portNumber);
			
		

			clientIn = null; // let gc() reclaim the space
			clientOut = null;
			client = null;
			avTransmitter= null;
			myBoss.clientDisconnected(clientNumber);
			// delete the record of this client
			clientEventHandler.clientDisconnected(clientNumber);
			ClientList.onlyList.deleteClient(clientNumber);
			System.out.println("Connection Closed Successfully");
			
			
		}
	}
	
	private boolean authorizeClient() throws IOException,ClassNotFoundException
	{
		
		final int pTryMax = 5;
//		 read user name and password from file
		sendMessage("+ Welcome to Buet Video Server\n          __|__\n   .____o(\"\")o____.\n        \"    '    \"\n");
		  

		
			sendMessage("+ Send your User name and Password to connect\n");
			user  = (String)clientIn.readObject();
			pass = (String)clientIn.readObject();
			clientEventHandler.clientMessage(client.getInetAddress().getHostAddress(),user+" "+pass);
			
			// add a authorization unit
			if(user==null || pass==null)return false;
			StringTokenizer t1  = new StringTokenizer(user);
			StringTokenizer t2  = new StringTokenizer(pass);
			if(t1.nextToken().equals("/user") && t2.nextToken().equals("/pass") && (t1.countTokens() ==1) && (t2.countTokens()==1))
			{
				 
					if((DatabaseManager.dataMan.searchPass(t1.nextToken())).equals( t2.nextToken()) )
						{			
								sendMessage("+ "+"Welcome "+user+"U r now authorized ");
								return true;
						}
					
			}
			
		sendMessage("- Wrong user name / password. ");
				
		
		
		return false;
		
		
	}
	private void sendMessage(String msg) throws IOException
	{
		clientOut.writeObject(msg);
		clientOut.flush();
		
	}
	
	private void processClient() throws IOException,ClassNotFoundException
	{
		
		// client is a valid client. ask him for usename and password
		// if he is valid, then increment authorized counter
		//and open a log for him
		if(!authorizeClient())return;
		
		
		//open a log recorder for this client
		// and grab an user number for him
		logClient = new LogRecorder(LogRecorder.LOG_CLIENT,client.getInetAddress().getHostAddress());
		logClient.addInfo("User: "+user+" Client from "+client.getInetAddress().getHostAddress());
		
		ClientList.onlyList.newClientAuthorized(clientNumber,user,client.getInetAddress().getHostAddress());
		
		myBoss.clientAuthorized(clientNumber);
		clientEventHandler.clientAuthorized(clientNumber);
		
		String msg,m1,m2,m3;
		currentActivity = IDLE;
		while(true)
		{
			msg = (String)clientIn.readObject();
			logClient.addInfo(msg+"\n");
			logClient.flush();
			if(msg == null)break;
			
			StringTokenizer t = new StringTokenizer(msg);
			m1 = t.nextToken(); // get the command

			// commands are in the format 
			/*
			 /<command> <parameter>
			  * Example:
			  	Command		Meaning
			  	/logout		Disconnect Client
			  	/media 1	Get the media number 1
			    /list		Give me the channel list 
			    /users		Give me the current users list
			 */
			if (m1.equals("/logout"))break;
			
			else if (m1.equals("/media"))
			{
				 
				if(myBoss.myMediaManager==null)
				{
					sendMessage("- Server not ready");
				}
				m2 = t.nextToken();
				if(m2==null)
				{
					sendMessage("- Syntax Error");
					continue;
				}
				//	if a transmit object was already create close refuse the connection
				if(avTransmitter !=null || rtpTransmitter!=null)
				{
					sendMessage("- Stop Existing one");
					continue;
				}
				
				if(m2.equals("capture"))
				{
					rtpTransmitter = new RTPTransmitter(this.client.getInetAddress().getHostAddress(),portNumber,TVCapture.getInstance().getDataSource());
					
					
					//TODO send the other ports here
					
					sendMessage("+ "+portNumber);
					currentActivity = WATCHING_TV;
					clientEventHandler.clientStateChanged(clientNumber,currentActivity);
					rtpTransmitter.start();
					
					
					continue; // be sure about this continue
				}
				
				int mediaId = Integer.parseInt(m2);
				
				Format fmt = null;
				
				
				
				// Create a transmit object with the specified params.
				
				avTransmitter = new AVTransmitter(myBoss.myMediaManager.getSource(mediaId),
								     client.getInetAddress().getHostAddress(), Integer.toString(portNumber), fmt);
				// Start the transmission
				String result = avTransmitter.doSetup();
				int nTrack = avTransmitter.getNofTracks();
				StringBuffer ports = new StringBuffer("+ ");
				for(int j=0;j<nTrack;j++)
				{
					ports.append(portNumber + 2*j );
					ports.append(" ");
				}
				System.out.println(ports.toString());
				sendMessage(ports.toString());
				currentActivity = WATCHING_VIDEO;
				//TODO send the other ports here
				clientEventHandler.clientStateChanged(clientNumber,currentActivity);
				avTransmitter.startTransmission();
				
				
				
				// result will be non-null if there was an error. The return
				// value is a String describing the possible error. Print it.
				if (result != null) {
				    System.err.println("Error : " + result);
				    
				}
				
			}
			/*
			else if(m1.equals("/tv"))
			{
				if(rtpTransmitter!=null || avTransmitter!=null)
				{
					sendMessage("- Stop Existing one");
					continue;
				}
				else
				{
					
				}
			}
			*/
			else if(m1.equals("/stop"))
			{
				if(currentActivity == WATCHING_VIDEO)
				{
					avTransmitter.stop();
					avTransmitter = null;
					currentActivity = IDLE;
					
					sendMessage("+ Stopped");
				}
				else if(currentActivity==WATCHING_TV)
				{
					rtpTransmitter.stop();
					rtpTransmitter = null; 
					currentActivity = IDLE;
					sendMessage("+ Stopped");
				}
				clientEventHandler.clientStateChanged(clientNumber,IDLE);
				
			}
			else if(m1.equals("/list"))
			{
				MediaManager mm = myBoss.myMediaManager;
				int cnt = myBoss.myMediaManager.getMediaCount();
				
				for(int i=0;i<cnt;i++)
				{
					sendMessage((String)mm.title.get(i));
					sendMessage(Integer.toString(i));
					sendMessage((String)mm.description.get(i));
					
				}
				sendMessage("Live Program");
				sendMessage("capture");
				sendMessage("Television Program :)");
				sendMessage("/###");
			}
			else 
			{
					sendMessage("- Unknown Command "+msg);
			}
			
			
			
		}
		
	}
	
public void run() {
		
		// Communicate with the client
		try{
			
			getStreams();
	
			// now listen what the client says;
			processClient();
			sendMessage("Good bye. and have a nice day :)");
			
		}
		
		catch(IOException e)
		{
			System.err.println("Error occured with Client about IO "+client.getInetAddress().getHostAddress());
			e.printStackTrace();
			
		}
		catch(ClassNotFoundException e)
		{
			System.err.println("Unknown Object Found");
			
		}
		
		
		finally{
			
			closeConnection();
			
			
		}
		
	}
}
